class AppConfHelper(object):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(AppConfHelper, cls).__new__(cls)
            cls.instance._appconf = None
        return cls.instance

    def initialize(self, config):
        self._appconf = config

    def find_replacement(self, key):
        parts = key.split('|')

        obj = self._appconf
        for x in range(0, len(parts)):
            part = parts[x]
            if part in obj:
                if x < len(parts):
                    obj = obj[part]
            else:
                obj = 'NLKF'
                break

        return obj



from . import BaseExpression, BaseIncludeExpression
import os
import json
from ..appconf import AppConfHelper
from ..utils import make_url_path


class IncludeExpression(BaseIncludeExpression):
    def __init__(self, settings):
        super(IncludeExpression, self).__init__(
            settings,
            settings.asset._settings.html.source,
            "_%s.html")

    @staticmethod
    def get_regex():
        return r"\[\%= include (?P<p_include_path>[a-zA-Z0-9_\-\\\/\.]+) \%\]"


class I18nExpression(BaseExpression):
    def __init__(self, settings):
        super(I18nExpression, self).__init__(settings)
        self._key = settings.match.group('p_i18n_key')

    def __call__(self, *args, **opts):
        return self.settings.asset._settings.i18n_helper.translate(self._key, self.settings.asset._lang)

    @staticmethod
    def get_regex_params():
        return ['p_i18n_key']

    @staticmethod
    def get_regex():
        return r"\[\%\- (?P<p_i18n_key>[a-zA-Z0-9_ \-]{1,50}(\:[a-zA-Z0-9_ \-]{1,50})*) \%\]"


class AppConfExpression(BaseExpression):
    def __init__(self, settings):
        super(AppConfExpression, self).__init__(settings)
        self._key = settings.match.group('p_appconf_key')
        self._filter = settings.match.group('p_appconf_filter')

    def __call__(self, *args, **opts):
        base = AppConfHelper().find_replacement(self._key)
        if self._filter == '':
            return str(base)
        elif self._filter == 'json':
            return json.dumps(base)
        else:
            return ''

    @staticmethod
    def get_regex_params():
        return ['p_appconf_key', 'p_appconf_filter']

    @staticmethod
    def get_regex():
        return r"\[\!(?P<p_appconf_key>[a-zA-Z0-9_\- ]{2,48}(\|[a-zA-Z0-9_\- ]{2,48})*)\!(?P<p_appconf_filter>[a-zA-Z0-9_\- ]{0,25})\]"


class StylesheetUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(StylesheetUrlExpression, self).__init__(settings)
        self._link_path = settings.match.group("p_stylesheet_url")
        self._dependency_path = os.path.join(
            self.settings.asset._settings.stylesheets.source,
            self._link_path)
        settings.asset.add_dependency(self._dependency_path)

    def __call__(self, *args, **opts):
        cache_entry = self.settings.asset._tool_cache.find_entry(
            self._dependency_path)
        if cache_entry:
            return make_url_path(
                self.settings.asset._settings.cdn_path,
                self.settings.asset._settings.cdn_url,
                cache_entry.target
            )
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_stylesheet_url"]

    @staticmethod
    def get_regex():
        return r"\[\%= stylesheet_url (?P<p_stylesheet_url>[a-zA-Z0-9_\-\\\/\.]+\.(css|sass|scss)) \%\]"


class ScriptUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(ScriptUrlExpression, self).__init__(settings)
        self._link_path = settings.match.group("p_script_url")
        self._dependency_path = os.path.join(
            self.settings.asset._settings.scripts.source,
            self._link_path)
        settings.asset.add_dependency(self._dependency_path)

    def __call__(self, *args, **opts):
        cache_entry = self.settings.asset._tool_cache.find_entry(self._dependency_path)
        if cache_entry:
            return make_url_path(
                self.settings.asset._settings.cdn_path,
                self.settings.asset._settings.cdn_url,
                cache_entry.target
            )
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_script_url"]

    @staticmethod
    def get_regex():
        return r"\[\%= script_url (?P<p_script_url>[a-zA-Z0-9_\-\\\/\.]+\.(js|coffee)) \%\]"


class ImageUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(ImageUrlExpression, self).__init__(settings)
        self._image_href = settings.match.group("p_image_href")
        self._dependency_path = os.path.join(
            self.settings.asset._settings.images.source,
            self._image_href)
        settings.asset.add_dependency(self._dependency_path)

    def __call__(self, *args, **opts):
        cache_entry = self.settings.asset._tool_cache.find_entry(self._dependency_path)
        if cache_entry:
            return make_url_path(
                self.settings.asset._settings.cdn_path,
                self.settings.asset._settings.cdn_url,
                cache_entry.target
            )
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_image_href"]

    @staticmethod
    def get_regex():
        return r"\[\%= image_url (?P<p_image_href>[a-zA-Z0-9_\-\\\/\.]+\.(png|jpg|gif)) \%\]"


class ResourceUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(ResourceUrlExpression, self).__init__(settings)
        self._key = settings.match.group('p_resource_url')

    def __call__(self, *args, **opts):
        return self.settings.asset._settings.resources.get_url(self._key)

    @staticmethod
    def get_regex_params():
        return ['p_resource_url']

    @staticmethod
    def get_regex():
        return r'\[\%= resource_url (?P<p_resource_url>[a-zA-Z0-9_\-]+((/[a-zA-Z0-9_\-]+)*)) \%\]'



from ..expressions import BaseExpression, BaseIncludeExpression
from ..cache import Cache
import os
from ..utils import make_url_path
from ..appconf import AppConfHelper
import simplejson


class IncludeExpression(BaseIncludeExpression):
    def __init__(self, settings):
        super(IncludeExpression, self).__init__(
            settings,
            settings.asset._settings.scripts.source,
            "_%s")

    @staticmethod
    def get_regex():
        return r"/\*= include (?P<p_include_path>[a-zA-Z0-9_\-\\\/\.]+\.(js|coffee)) \*/"


class ScriptUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(ScriptUrlExpression, self).__init__(settings)
        self._script_path = settings.match.group("p_script_url")
        self._dependency_path = os.path.join(
            settings.asset._settings.scripts.source,
            self._script_path)
        settings.asset.add_dependency(self._dependency_path)

    def __call__(self, **opts):
        tool_cache = Cache()
        cache_entry = tool_cache.find_entry(self._dependency_path)
        if cache_entry:
            return make_url_path(
                self.settings.asset._settings.cdn_path,
                self.settings.asset._settings.cdn_url,
                cache_entry.target
            )
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_script_url"]

    @staticmethod
    def get_regex():
        return r"/\*= script_url (?P<p_script_url>[a-zA-Z0-9_\-\\\/\.]+\.(js|coffee)) \*/"


class AppConfExpression(BaseExpression):
    def __init__(self, settings):
        super(AppConfExpression, self).__init__(settings)
        self._key = settings.match.group('p_appconf_key')
        self._filter = settings.match.group('p_appconf_filter')

    def __call__(self, *args, **opts):
        base = AppConfHelper().find_replacement(self._key)
        if self._filter is None:
            return str(base)
        elif self._filter == 'json':
            return simplejson.dumps(base)
        return None

    @staticmethod
    def get_regex_params():
        return ['p_appconf_key', 'p_appconf_filter']

    @staticmethod
    def get_regex():
        return r"/\*= config (?P<p_appconf_key>[a-zA-Z0-9_\- ]{2,48}(\|[a-zA-Z0-9_\- ]{2,48})*)((\:(?P<p_appconf_filter>[a-zA-Z0-9]{2,20}))?) \*/"


class ResourceUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(ResourceUrlExpression, self).__init__(settings)
        self._key = settings.match.group('p_resource_url')

    def __call__(self, *args, **opts):
        return self.settings.asset._settings.resources.get_url(self._key)

    @staticmethod
    def get_regex_params():
        return ['p_resource_url']

    @staticmethod
    def get_regex():
        return r'/\*= resource_url (?P<p_resource_url>[a-zA-Z0-9_\-]+((/[a-zA-Z0-9_\-]+)*)) \*/'



import io
import yaml


class I18nHelper(object):
    def __init__(self, data):
        self.translations = {}
        for lang, file_path in data.iteritems():
            self.load_translation(lang, file_path)

    def load_translation(self, lang, file_path):
        with io.open(file_path, 'r', encoding='utf8') as f:
            self.translations[lang] = yaml.load(f)

    def translate(self, key, lang):
        parts = key.split(':')

        obj = None
        if not lang in self.translations:
            obj = self.translations['en']
        else:
            obj = self.translations[lang]

        for x in range(0, len(parts)):
            part = parts[x]
            if part in obj:
                obj = obj[part]
            else:
                obj = 'NLKF'
                break

        return obj



from .db import Model
from sqlalchemy import Column, Integer, String, DateTime
import os
import datetime
from .utils import get_file_hash


class CacheEntry(Model):
    __tablename__ = "cache"
    id = Column(Integer, primary_key=True)
    source = Column(String(512))
    target = Column(String(512))
    lang = Column(String(10), nullable=True)
    last_modified = Column(DateTime)
    checksum = Column(String(64))

    def __init__(self, source, target, lang=None):
        self.source = source
        self.target = target
        self.lang = lang
        self.update_last_modified()
        self.update_checksum()

    def update_checksum(self):
        self.checksum = get_file_hash(self.source)

    def update_last_modified(self):
        self.last_modified = datetime.datetime.fromtimestamp(
            os.path.getmtime(self.source))

    def __repr__(self):
        return "%d - s:'%s', t:'%s', m:'%s', c:'%s'" % (
            self.id, self.source, self.target, self.last_modified,
            self.checksum)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def file_modified(self):
        if not os.path.exists(self.target):
            return True
        source_last_mod = datetime.datetime.fromtimestamp(
            os.path.getmtime(self.source))
        if source_last_mod > self.last_modified:
            return True
        return False



from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from functools import wraps
import os

db_session = None

Model = declarative_base()


def entry_point(func):
    @wraps(func)
    def internal(*args, **kwargs):
        result = None
        cache_db_path = os.path.join(args[0].cache_path, "cache.db")
        engine = create_engine("sqlite:///" + cache_db_path,
                               convert_unicode=True)
        global db_session
        db_session = scoped_session(sessionmaker(autocommit=False,
                                                 autoflush=False,
                                                 bind=engine))
        global Model
        Model.query = db_session.query_property()
        from . import models
        Model.metadata.create_all(bind=engine)
        try:
            result = func(*args, **kwargs)
        finally:
            db_session.remove()
        return result

    return internal



from hashlib import sha256
import io
import os
import urlparse


def shorten_digest(hexdigest):
    alphabet = list("0123456789abcdefghijklmnopqrstuvwxyz")
    alpha_len = len(alphabet)

    unique_id = int(hexdigest, 16)
    output = ""
    while unique_id:
        unique_id, digit = divmod(unique_id, alpha_len)
        output += alphabet[digit]
    return output


def get_file_hash(path, unique=False):
    hash = sha256()
    if unique:
        hash.update(os.urandom(32))
    with io.open(path, 'rb') as f:
        for chunk in iter(lambda: f.read(io.DEFAULT_BUFFER_SIZE), b''):
            hash.update(chunk)
    return shorten_digest(hash.hexdigest())


def get_data_hash(data):
    hash = sha256()
    hash.update(data)
    return hash.hexdigest()


def load_file(path):
    data = None
    with io.open(path, 'r', encoding='utf-8') as f:
        data = f.read()
    return data


def save_file(path, data):
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    with io.open(path, 'w', encoding='utf-8') as f:
        f.write(data)


def make_url_path(base_folder, base_url, path):
    prefix = os.path.commonprefix([
        base_folder,
        path
    ])
    if prefix == base_folder:
        path_part = path[len(prefix)+1:].replace('\\', '/')
        return urlparse.urljoin(base_url, path_part)
    return ''



from ..expressions import BaseExpression, BaseIncludeExpression
import os
from ..cache import Cache
from ..utils import make_url_path


class IncludeExpression(BaseIncludeExpression):
    def __init__(self, settings):
        super(IncludeExpression, self).__init__(
            settings,
            settings.asset._settings.stylesheets.source,
            "_%s")

    @staticmethod
    def get_regex():
        return r"/\*= include (?P<p_include_path>[a-zA-Z0-9_\-\\\/\.]+\.(css|scss|sass)) \*/"


class ImageUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(ImageUrlExpression, self).__init__(settings)
        self._image_path = settings.match.group("p_image_url")
        self._dependency_path = os.path.join(
            settings.asset._settings.images.source,
            self._image_path)
        settings.asset.add_dependency(self._dependency_path)

    def __call__(self, **opts):
        tool_cache = Cache()
        cache_entry = tool_cache.find_entry(self._dependency_path)
        if cache_entry:
            return make_url_path(
                self.settings.asset._settings.cdn_path,
                self.settings.asset._settings.cdn_url,
                cache_entry.target
            )
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_image_url"]

    @staticmethod
    def get_regex():
        return r"/\*= image_url (?P<p_image_url>[a-zA-Z0-9_\-\\\/\.]+\.(png|gif|jpg)) \*/"


class FontUrlExpression(BaseExpression):
    def __init__(self, settings):
        super(FontUrlExpression, self).__init__(settings)
        self._font_path = settings.match.group("p_font_url")
        self._dependency_path = os.path.join(
            settings.asset._settings.fonts.source,
            self._font_path)
        settings.asset.add_dependency(self._dependency_path)

    def __call__(self, **opts):
        tool_cache = Cache()
        cache_entry = tool_cache.find_entry(self._dependency_path)
        if cache_entry:
            return make_url_path(
                self.settings.asset._settings.cdn_path,
                self.settings.asset._settings.cdn_url,
                cache_entry.target
            )
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_font_url"]

    @staticmethod
    def get_regex():
        return r"/\*= font_url (?P<p_font_url>[a-zA-Z0-9_\-\\\/\.]+\.(ttf|svg|woff|eot)) \*/"



import re
from .expressions import ExpressionSettings


def get_match_expression_class(match_obj, classes):
    for class_name in classes:
        matches = False
        for param in class_name.get_regex_params():
            # one match is enough
            if match_obj.group(param) is not None:
                matches = True
        if matches:
            return class_name
    return None


def get_expressions_regex(classes):
    grouped = ["(%s)" % x.get_regex() for x in classes]
    return "|".join(grouped)


class ExpressionProcessor(object):
    def __init__(self, asset, resolvers):
        self._asset = asset
        self._data = asset._data
        self._resolvers = resolvers
        self._expressions = []

    def parse(self):
        group_regex = get_expressions_regex(self._resolvers)
        for match in re.finditer(group_regex, self._data):
            class_name = get_match_expression_class(
                match, self._resolvers)

            if class_name is not None:
                expr = class_name(ExpressionSettings(
                    self, self._asset, match))
                self._expressions.append(expr)

    def compile(self, settings, path):
        result = ''
        start = 0
        end = len(self._data)
        for expression in self._expressions:
            span = expression.settings.match.span()
            end = span[0]

            result += self._data[start:end]
            expr_result = expression(settings=settings, path=path)

            result += expr_result
            start = span[1]
            end = len(self._data)

        result += self._data[start:end]
        self._asset._data = result



import os


def detour_directory(path, func):
    for object_name in os.listdir(path):
        full_path = os.path.join(path, object_name)
        if os.path.isfile(full_path):
            func(full_path)
        elif os.path.isdir(full_path):
            detour_directory(full_path, func)



from .models import CacheEntry
from . import db
import os


class Cache(object):
    def __new__(cls, commit=False):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Cache, cls).__new__(cls)
            cls.instance.entries = CacheEntry.query.all()
        return cls.instance

    def check(self):
        for entry in self.entries:
            if not os.path.exists(entry.source):
                if os.path.exists(entry.target):
                    os.remove(entry.target)
                db.db_session.delete(entry)

        db.db_session.commit()

    def find_entry(self, source, lang=None):
        for entry in self.entries:
            if entry.source == source and entry.lang == lang:
                return entry
        return None

    def add(self, entry):
        self.entries.append(entry)
        db.db_session.add(entry)
        db.db_session.commit()

    def update(self, entry):
        entry.update_last_modified()
        entry.update_checksum()
        db.db_session.commit()



import os
from ..utils import load_file


class ExpressionSettings(object):
    def __init__(self, processor, asset, match):
        self._processor = processor
        self._asset = asset
        self._match = match

    @property
    def processor(self):
        return self._processor

    @property
    def asset(self):
        return self._asset

    @property
    def match(self):
        return self._match


class BaseExpression(object):
    def __init__(self, settings):
        self._settings = settings

    @property
    def settings(self):
        return self._settings


class BaseIncludeExpression(BaseExpression):
    def __init__(self, settings, include_from, include_pattern,
                 param="p_include_path"):
        super(BaseIncludeExpression, self).__init__(settings)
        self._include_path = self.settings.match.group(param)
        self._dependency_path = os.path.join(
            include_from, self._include_path)
        self._dependency_path = os.path.join(
            os.path.dirname(self._dependency_path),
            include_pattern % os.path.basename(self._dependency_path)
        )
        self.settings.asset.add_dependency(
            self._dependency_path,
            self.settings.asset._lang)

    def __call__(self, **opts):
        cache_entry = self.settings.asset._tool_cache.find_entry(
            self._dependency_path,
            self.settings.asset._lang)
        if cache_entry:
            return load_file(cache_entry.target)
        return ""

    @staticmethod
    def get_regex_params():
        return ["p_include_path"]





class ResourceSet(object):
    def __init__(self, settings):
        self.resources = self.parse_resource('', '', settings)

    @classmethod
    def urljoin(cls, *args):
        if len(args) > 0 and len(args[0]) == 0:
            args = args[1:]
        return '/'.join(map(lambda x: x.rstrip('/'), args))

    @classmethod
    def parse_resource(cls, base_alias, base_url, settings):
        result = {}
        resource_base = cls.urljoin(base_url, settings['base_url'])
        for alias, value in settings['resources'].iteritems():
            cur_base = cls.urljoin(base_alias, alias)
            if type(value) == dict:
                result.update(cls.parse_resource(
                    cur_base, resource_base, value))
            else:
                result[cur_base] = cls.urljoin(resource_base, value)

        return result

    def __repr__(self):
        r = 'RESOURCES:\r\n'
        for alias, url in self.resources.iteritems():
            r += '{alias}\t{url}\r\n'.format(alias=alias, url=url)
        return r

    def get_url(self, alias):
        if alias in self.resources:
            return self.resources[alias]
        return '#'



import os
from .cache import Cache
from .models import CacheEntry
from .utils import get_file_hash, save_file, load_file
import shutil
import io
from .compiler import ExpressionProcessor
from .expressions import stylesheets, scripts, html
import subprocess
import tempfile


class AssetCollection(object):
    def __init__(self, file_list, settings):
        self._assets = []
        self._settings = settings
        for path in file_list:
            res = get_asset_objects(path, settings)
            if type(res) is list:
                for asset in res:
                    self._assets.append(asset)
                    self._assets[-1]._collection = self
                    self._assets[-1]._settings = settings
            else:
                if res is None:
                    continue
                self._assets.append(res)
                self._assets[-1]._collection = self
                self._assets[-1]._settings = settings

    def find_asset(self, path, lang):
        for asset in self._assets:
            if asset._path == path and asset._lang == lang:
                return asset
        return None

    def pick_dependencies(self):
        print('Found {count:d} assets'.format(count=len(self._assets)))

        if self._settings.verbose:
            print("Picking dependencies...")
        for asset in self._assets:
            asset.parse()
            if self._settings.verbose:
                print(asset)
                print('Dependencies {dependencies}\n'.format(
                      dependencies=asset._dependencies))

        self._assets = DependencyResolver.topological_sort(self._assets)
        if self._settings.verbose:
            print('Build order:\n{collection}\n'.format(
                  collection=self._assets))

    def build(self):
        print('Building assets...')
        for asset in self._assets:
            asset.compile(force=self._settings.force)
        print('Build done.')


class DependencyResolver(object):
    @staticmethod
    def topological_sort(assets_unsorted):
        assets_sorted = []

        while len(assets_unsorted) > 0:
            acyclic = False
            for asset in assets_unsorted:
                for dependency in asset._dependencies:
                    if dependency in assets_unsorted:
                        break
                else:
                    acyclic = True
                    assets_unsorted.remove(asset)
                    assets_sorted.append(asset)

            if not acyclic:
                raise RuntimeError('A cyclic dependency occurred')

        return assets_sorted


class Asset(object):
    FILE = 0
    STRING = 1

    def __init__(self, resource_type, path, lang):
        self._resource_type = resource_type
        self._path = path
        self._lang = lang
        self._collection = None
        self._settings = None
        self._dependencies = []
        self._tool_cache = Cache()
        self._flag_modified = False

    def is_partial(self, path):
        return os.path.basename(path).startswith("_")

    def get_target_path(self, **opts):
        common_prefix = os.path.commonprefix([
            self._path,
            self._get_source_dir()])
        path_part = self._path[len(common_prefix)+1:]
        if 'hash' in opts:
            parts = os.path.splitext(path_part)
            new_filename = '%s-%s' % (parts[0], opts['hash'])
            path_part = '%s%s' % (new_filename, parts[1])
        if 'change_extension' in opts:
            new_ext = opts['change_extension']
            parts = os.path.splitext(path_part)
            path_part = '%s%s' % (parts[0], new_ext)
        if 'lang' in opts and not(opts['lang'] is None):
            lang = opts['lang']
            parts = os.path.splitext(path_part)
            path_part = '%s-%s%s' % (parts[0], lang, parts[1])
        if self.is_partial(path_part):
            target_path = os.path.join(self._get_partials_dir(), path_part)
        else:
            target_path = os.path.join(self._get_target_dir(), path_part)
        return target_path

    def __repr__(self):
        if self._lang is None:
            t = '{path}'
        else:
            t = '{path} ({lang})'
        common_prefix = os.path.commonprefix([
            self._path,
            self._get_source_dir()])
        return t.format(path=self._path[len(common_prefix) + 1:],
                        lang=self._lang)

    def add_dependency(self, path, lang=None):
        dependency = self._collection.find_asset(path, lang)
        if dependency:
            if dependency not in self._dependencies:
                self._dependencies.append(dependency)
        else:
            print("Couldn't find dependency with path %s" % path)

    def __eq__(self, other):
        return self._path == other._path and self._lang == other._lang

    def __ne__(self, other):
        return self._path != other._path and self._lang != other._lang

    def parse(self):
        self._parse()

    def dependencies_modified(self):
        for dep_asset in self._dependencies:
            if dep_asset._flag_modified:
                return True
        return False

    def compile(self, force=False):
        if self._resource_type == Asset.FILE:
            cache_entry = self._tool_cache.find_entry(self._path, self._lang)

            file_modified = True if cache_entry is None\
                else cache_entry.file_modified() or self.dependencies_modified()

            if file_modified or force:
                if cache_entry:
                    if os.path.exists(cache_entry.target):
                        os.remove(cache_entry.target)

                target_path = self._get_target_path()
                self._compile(target_path)

                if cache_entry:
                    cache_entry.target = target_path
                    self._tool_cache.update(cache_entry)
                    print('Updated {asset}'.format(asset=self))
                else:
                    cache_entry = CacheEntry(self._path, target_path, self._lang)
                    self._tool_cache.add(cache_entry)
                    print('Created {asset}'.format(asset=self))
                self._flag_modified = True
            else:
                if self._settings.verbose:
                    print('Cached {asset}'.format(asset=self))
        else:
            print("String asset")


class StringAsset(Asset):
    def __init__(self, path, lang=None):
        super(StringAsset, self).__init__(Asset.STRING, path, lang)
        self._data = None


class TextAsset(Asset):
    def __init__(self, path, lang=None):
        super(TextAsset, self).__init__(Asset.FILE, path, lang)
        self._data = None

        split = os.path.splitext(path)
        self._basename = split[0]
        self._extension = split[1]

    def load(self):
        with io.open(self._path, 'r', encoding='utf-8') as f:
            self._data = f.read()

    def save(self, path):
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        save_file(path, self._data)


class StylesheetAsset(TextAsset):
    @staticmethod
    def supported_extensions():
        return ['.css', '.scss']

    @staticmethod
    def get_languages(settings):
        return settings.stylesheets.languages

    def _get_partials_dir(self):
        return os.path.join(self._settings.partials, 'stylesheets')

    def _get_source_dir(self):
        return self._settings.stylesheets.source

    def _get_target_dir(self):
        return self._settings.stylesheets.target

    def _get_target_path(self):
        return self.get_target_path(hash=get_file_hash(self._path, unique=True))

    def _parse(self):
        self.load()
        self._processor = ExpressionProcessor(self, [
            stylesheets.ImageUrlExpression,
            stylesheets.IncludeExpression,
            stylesheets.FontUrlExpression
        ])
        self._processor.parse()

    def minify(self):
        temp_path = tempfile.mkdtemp()

        source_file = os.path.join(temp_path, "source.css")
        save_file(source_file, self._data)
        target_file = os.path.join(temp_path, "target.css")

        proc = subprocess.Popen(
            [
                "java",
                "-Xss100m",
                "-jar",
                self._settings.yuicompressor_file,
                "--type",
                "css",
                "-o",
                target_file,
                source_file
            ],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        out, err = proc.communicate()

        self._data = load_file(target_file)
        shutil.rmtree(temp_path)

    def _compile(self, target_path):
        self._processor.compile(self._settings, target_path)
        if self._settings.minify and not self.is_partial(target_path):
            if self._settings.verbose:
                print('Minifying {asset}'.format(asset=self))
            self.minify()
        self.save(target_path)


class ScriptAsset(TextAsset):
    @staticmethod
    def supported_extensions():
        return ['.js', '.coffee']

    @staticmethod
    def get_languages(settings):
        return settings.scripts.languages

    def _get_partials_dir(self):
        return os.path.join(self._settings.partials, 'scripts')

    def _get_source_dir(self):
        return self._settings.scripts.source

    def _get_target_dir(self):
        return self._settings.scripts.target

    def _get_target_path(self):
        return self.get_target_path(
            hash=get_file_hash(self._path, unique=True),
            change_extension='.js'
        )

    def _parse(self):
        self.load()
        self._processor = ExpressionProcessor(self, [
            scripts.IncludeExpression,
            scripts.ScriptUrlExpression,
            scripts.AppConfExpression,
            scripts.ResourceUrlExpression
        ])
        self._processor.parse()

    def minify(self):
        temp_path = tempfile.mkdtemp()

        source_file = os.path.join(temp_path, "source.js")
        save_file(source_file, self._data)
        target_file = os.path.join(temp_path, "target.js")

        proc = subprocess.Popen(
            [
                "java",
                "-jar",
                self._settings.yuicompressor_file,
                "--type",
                "js",
                "-o",
                target_file,
                source_file
            ],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        out, err = proc.communicate()

        self._data = load_file(target_file)
        shutil.rmtree(temp_path)

    def compile_coffee(self):
        temp_path = tempfile.mkdtemp()

        source_file = os.path.join(temp_path, "source.coffee")
        save_file(source_file, self._data)
        target_file = os.path.join(temp_path, "source.js")

        proc = subprocess.Popen(
            [
                "coffee",
                "-c",
                source_file
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        out, err = proc.communicate()

        self._data = load_file(target_file)
        shutil.rmtree(temp_path)

    def _compile(self, target_path):
        self._processor.compile(self._settings, target_path)
        if self._extension == '.coffee':
            if self._settings.verbose:
                print('Using CoffeeScript Compiler for {asset}'.format(asset=self))
            self.compile_coffee()
        if self._settings.minify and not self.is_partial(target_path):
            if self._settings.verbose:
                print('Minifying {asset}'.format(asset=self))
            self.minify()
        self.save(target_path)


class HtmlAsset(TextAsset):
    @staticmethod
    def supported_extensions():
        return ['.html']

    @staticmethod
    def get_languages(settings):
        return settings.html.languages

    def _get_partials_dir(self):
        return os.path.join(self._settings.partials, 'html')

    def _get_source_dir(self):
        return self._settings.html.source

    def _get_target_dir(self):
        return self._settings.html.target

    def _get_target_path(self):
        return self.get_target_path(lang=self._lang)

    def _parse(self):
        self.load()
        self._processor = ExpressionProcessor(self, [
            html.IncludeExpression,
            html.StylesheetUrlExpression,
            html.ScriptUrlExpression,
            html.ImageUrlExpression,
            html.AppConfExpression,
            html.I18nExpression,
            html.ResourceUrlExpression
        ])
        self._processor.parse()

    def minify(self):
        temp_path = tempfile.mkdtemp()

        source_file = os.path.join(temp_path, "source.html")
        save_file(source_file, self._data)
        target_file = os.path.join(temp_path, "target.html")

        proc = subprocess.Popen(
            [
                "java",
                "-jar",
                self._settings.htmlcompressor_file,
                "--type",
                "html",
                "--mask",
                "*.html",
                "-o",
                target_file,
                source_file,
                "--remove-intertag-spaces"
            ],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE
        )
        out, err = proc.communicate()

        self._data = load_file(target_file)
        shutil.rmtree(temp_path)

    def _compile(self, target_path):
        self._processor.compile(self._settings, target_path)
        if self._settings.minify and not self.is_partial(target_path):
            if self._settings.verbose:
                print('Minifying {asset}'.format(asset=self))
            self.minify()
        self.save(target_path)


class BinaryAsset(Asset):
    def __init__(self, path, lang=None):
        super(BinaryAsset, self).__init__(Asset.FILE, path, lang)

    def _get_target_path(self):
        return self.get_target_path(hash=get_file_hash(self._path, unique=True))

    def _parse(self):
        pass

    def _compile(self, target_path):
        if not os.path.exists(os.path.dirname(target_path)):
            os.makedirs(os.path.dirname(target_path))
        shutil.copy(self._path, target_path)


class ImageAsset(BinaryAsset):
    def __init__(self, path, lang=None):
        super(ImageAsset, self).__init__(path, lang)

    @staticmethod
    def supported_extensions():
        return ['.png', '.jpg', '.gif']

    @staticmethod
    def get_languages(settings):
        return settings.images.languages

    def _get_partials_dir(self):
        return os.path.join(self._settings.partials, 'images')

    def _get_source_dir(self):
        return self._settings.images.source

    def _get_target_dir(self):
        return self._settings.images.target


class FontAsset(BinaryAsset):
    def __init__(self, path, lang=None):
        super(FontAsset, self).__init__(path, lang)

    @staticmethod
    def supported_extensions():
        return ['.eot', '.svg', '.ttf', '.woff']

    @staticmethod
    def get_languages(settings):
        return settings.fonts.languages

    def _get_partials_dir(self):
        return os.path.join(self._settings.partials, 'fonts')

    def _get_source_dir(self):
        return self._settings.fonts.source

    def _get_target_dir(self):
        return self._settings.fonts.target


def get_asset_objects(path, settings):
    asset_classes = [
        ImageAsset,
        FontAsset,
        StylesheetAsset,
        HtmlAsset,
        ScriptAsset
    ]

    file_ext = os.path.splitext(path)[1]
    for asset_class in asset_classes:
        if file_ext in asset_class.supported_extensions():
            langs = asset_class.get_languages(settings)
            if langs is None:
                return asset_class(path, None)
            else:
                return [asset_class(path, lang) for lang in langs]

    return None



from setuptools import setup, find_packages


setup(
    name='assetoolz',
    version='0.0.4',
    description='Web assets build system',
    author='Alexander Pyatkin',
    author_email='asp@thexyz.net',
    packages=find_packages('.'),
    install_requires=[
        'setuptools>=5.4',
        'sqlalchemy>=0.9,<1.0',
        'pyyaml>=3.11',
        'simplejson>=3.6,<4.0'
    ],
    url='https://github.com/aspyatkin/assetoolz',
    license='MIT',
    entry_points=dict(
        console_scripts=['assetoolz = assetoolz:main'])
)



from __future__ import absolute_import
import argparse
import os
from .detour import detour_directory
from .cache import Cache
from .db import entry_point
from .appconf import AppConfHelper
from .assets import AssetCollection
import yaml
from .utils import load_file
from .i18n import I18nHelper
from .resource import ResourceSet


class AssetSettings:
    def __init__(self, data):
        self._data = data

    @property
    def source(self):
        return self._data['source']

    @property
    def target(self):
        return self._data['target']

    @property
    def languages(self):
        return None


class LocalizedAssetSettings(AssetSettings):
    @property
    def languages(self):
        return self._data['languages']


class Settings:
    def __init__(self, conf_file, verbose, force):
        self._data = yaml.load(load_file(conf_file))
        self._html = LocalizedAssetSettings(self._data['html'])
        self._images = AssetSettings(self._data['images'])
        self._fonts = AssetSettings(self._data['fonts'])
        self._scripts = AssetSettings(self._data['scripts'])
        self._stylesheets = AssetSettings(self._data['stylesheets'])

        self._i18n_helper = I18nHelper(self._data['i18n'])
        self._resources = ResourceSet(self._data['resource'])

        self._verbose = verbose
        self._force = force

    @property
    def verbose(self):
        return self._verbose

    @property
    def cdn_path(self):
        return self._data['cdn']['path']

    @property
    def cdn_url(self):
        return self._data['cdn']['url']

    @property
    def force(self):
        return self._force

    @property
    def minify(self):
        return self._data['minify']

    @property
    def yuicompressor_file(self):
        return self._data['yuicompressor_file']

    @property
    def htmlcompressor_file(self):
        return self._data['htmlcompressor_file']

    @property
    def appconf(self):
        return self._data['config']

    @property
    def i18n(self):
        return self._data['i18n']

    @property
    def cache_path(self):
        return self._data['cache']

    @property
    def partials(self):
        return os.path.join(self.cache_path, "partials")

    @property
    def html(self):
        return self._html

    @property
    def images(self):
        return self._images

    @property
    def fonts(self):
        return self._fonts

    @property
    def scripts(self):
        return self._scripts

    @property
    def stylesheets(self):
        return self._stylesheets

    @property
    def i18n_helper(self):
        return self._i18n_helper

    @property
    def resources(self):
        return self._resources


@entry_point
def compile(settings):
    file_list = []
    update_file_list = lambda x: file_list.append(x)

    detour_directory(settings.html.source, update_file_list)
    detour_directory(settings.scripts.source, update_file_list)
    detour_directory(settings.stylesheets.source, update_file_list)
    detour_directory(settings.images.source, update_file_list)
    detour_directory(settings.fonts.source, update_file_list)

    tool_cache = Cache()
    tool_cache.check()

    assets = AssetCollection(file_list, settings)
    assets.pick_dependencies()
    assets.build()


def main():
    parser = argparse.ArgumentParser(description="assetoolz")
    parser.add_argument('config', metavar='config', type=str)
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-f', '--force', action='store_true')
    args = parser.parse_args()

    config = args.config
    verbose = args.verbose
    force = args.force
    settings = Settings(config, verbose, force)
    AppConfHelper().initialize(settings.appconf)
    compile(settings)


if __name__ == '__main__':
    main()
